/***************************************************************************
                          cdb.cpp  -  description
                             -------------------
    begin                : Thu Jan 23 2003
    copyright            : (C) 2003 by Oleksiy Kohany
    email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *                                                                         *
 *   class CDb - implements basic operations on multiple sequence file:    *
 *   - searching particular locus                                          *
 *   - extracting subsequence from lucus rapidly, assumes that             *
 *     all data lines are of the same length                               *
 *                                                                         *
 ***************************************************************************/
#include "cdb.h"
#include "constants.h"
#include <iostream>
#include <string.h>

//#define DEBUG

#include <sys/types.h>
#include <sys/stat.h>

ostream& operator << (ostream& os, CLP& lp) {
  os << lp.fname << "\t" << lp.headPos << "\t" << lp.dataPos << "\n";
  return os;
}

/** ==================================================== */
CDb::CDb(const char* fn) {
   ifs = 0;
   if(fn == 0) {
      throw *(new badfilename_exception(fn));
   }
#ifdef IOSTREAM
   ifs = new ifstream;
#endif
   sDbName = sDbDir = fn;
   sDbDir.erase(sDbDir.rfind('/') + 1);
   sDbName.erase(0, sDbDir.rfind('/') + 1);
   if(loadIdx(fn)) {
#ifdef DEBUG
      cerr << "close_ifs("<<sDbName<<") in CDb::CDb("<<fn<<");\n";
#endif
      sIDX = fn;
   }
   else {
      if(open_ifs(sDbName.c_str()) == 0) {
         throw *(new badfilename_exception(sDbName.c_str()));
      }
      sIDX.assign(fn);
      sIDX.append(".idx");
      int size = loadIndexUpToDate(fn, sIDX.c_str());
#ifdef DEBUG
      cerr << "loadIndexUpToDate(" << fn << ", " << sIDX << ") = " << size << "\n";
#endif
   }
   iIDXSize = lpmap.size();
   iLastPos = 0;
}
/** ==================================================== */
CDb::~CDb(){
#ifdef DEBUG
   cerr << "close_ifs("<<sDbName<<") in CDb::~CDb();\n";
#endif
   close_ifs();
#ifdef IOSTREAM
   delete ifs;
#endif
   if(lpmap.size() > 0 && iIDXSize < lpmap.size()) {
      saveIdx(sIDX.c_str());
   }
}
/** ==================================================== */
long CDb::getHeadPos(const char* locus){
  if(findLocus(locus)) {
    return lpmap[locus].headPos;
  }
  return -1;
}
/** ==================================================== */
long CDb::getDataPos(const char* locus){
  if(findLocus(locus)) {
    return lpmap[locus].dataPos;
  }
  return -1;
}
/** ==================================================== */
#ifdef IOSTREAM
ifstream* CDb::getHead(const char* locus){
#else
FILE* CDb::getHead(const char* locus){
#endif
  if(findLocus(locus)) {
#ifdef IOSTREAM
    ifs->clear();
    ifs->seekg(lpmap[locus].headPos, ios::beg);
    return ifs;
#else
    fseek(ifs, lpmap[locus].headPos, SEEK_SET);
    return ifs;
#endif
  }
  return 0;
}
/** ==================================================== */
#ifdef IOSTREAM
ifstream* CDb::getData(const char* locus){
#else
FILE* CDb::getData(const char* locus){
#endif
   if(findLocus(locus)) {
#ifdef IOSTREAM
#ifdef DEBUG
      cerr << "Seeking " << lpmap[locus].dataPos << " ... ";
#endif
      ifs->clear();
      ifs->seekg(lpmap[locus].dataPos, ios::beg);
#ifdef DEBUG
      cerr << ifs->tellg() << "\n";
#endif
#else
      fseek(ifs, lpmap[locus].dataPos, SEEK_SET);
#endif
      return ifs;
   }
   return 0;
}
#ifdef IOSTREAM
/** ==================================================== */
int CDb::findLocus(const char* locus){
   if(lpmap.find(locus) != lpmap.end()) {
      if(sDbName != lpmap[locus].fname) {
#ifdef DEBUG
         cerr << "close_ifs("<<sDbName<<") in int CDb::findLocus("<<locus<<"); \n";
#endif
         close_ifs();
         if(open_ifs(lpmap[locus].fname.c_str()) == 0) {
           return 0;
         }
         iLastPos = 0;
      }
      if(iIDXSize == 0) {
        return 1;
      }
      if(cmpage((sDbDir + lpmap[locus].fname).c_str(), sIDX.c_str()) < 0) {
#ifdef DEBUG
        cerr << sIDX << " is younger than " << lpmap[locus].fname << "\n";
#endif
        return 1;
      }
      else {

#ifdef DEBUG
        cerr << sIDX << " is older than " << lpmap[locus].fname << "\n";
#endif
      }
   }
#ifdef DEBUG
  cerr << "IOSTREAM: '" << locus << "' Not found in index. Tracking. iLastPos= " << iLastPos << "\n";
  int iCnt = 0;
#endif
   long iHP;
   long iDP;
   iHP = iLastPos;
   iDP = iLastPos;
   char sLine[MAXLINELENGTH];
//   string sLine;
   char sName[MAXLINELENGTH];
   if(ifs->is_open()) {
      ifs->clear();
      ifs->seekg(iLastPos, ios::beg);
      while(!ifs->getline(sLine, MAXLINELENGTH-1).eof()) {
//      while(getline(*ifs, sLine) && !ifs->fail()) {
#ifdef DEBUG
      cerr << "'" << sLine << "'\n";
#endif
         ifs->clear();
         sLine[MAXLINELENGTH-1] = '\0';
         if(sLine[0] == ';' || sLine[0] == '>') {
            switch(sLine[0]) {
               case '>':
                  sscanf(sLine + 1, "%s", sName);
                  break;
               case ';':
                  while(!ifs->getline(sLine, MAXLINELENGTH).eof() && sLine[0] == ';' ) {
                    ifs->clear();
                  }
//                  while(getline(*ifs, sLine) && !ifs->fail() && sLine[0] == ';');
                  sscanf(sLine, "%s", sName);
                  break;
            }
            iLastPos = iDP = ifs->tellg();
            ifs->clear();
            lpmap[sName].assign(sDbName, iHP, iDP);
#ifdef DEBUG
            iCnt++;
            cerr << sName << "<=>" << locus << "\n";
#endif
            if(strcmp(sName, locus) == 0) {
               return 1;
            }
         }
         iHP = iLastPos = ifs->tellg();
      }
#ifdef DEBUG
      cerr << "Failed ? " << ifs->fail() << "\n";
      cerr << "'" << sLine << "'\n";
      cerr << "Scanned " << iCnt << " locies\n";
#endif
   }
   ifs->clear();
   return 0;
}
#else
/** ==================================================== */
int CDb::findLocus(const char* locus){
// cerr << "Searching for " << locus << "\n";
   if(lpmap.find(locus) != lpmap.end()) {
      if(sDbName != lpmap[locus].fname) {
         close_ifs();
         if(open_ifs(lpmap[locus].fname.c_str()) == 0) {
           return 0;
         }
         iLastPos = 0;
         if(iIDXSize == 0) {
           return 1;
         }
         if(cmpage((sDbDir + lpmap[locus].fname).c_str(), sIDX.c_str()) < 0) {
#ifdef DEBUG
           cerr << sIDX << " is younger than " << lpmap[locus].fname << "\n";
#endif
           return 1;
         }
         else {
#ifdef DEBUG
           cerr << sIDX << " is older than " << lpmap[locus].fname << "\n";
#endif
         }
      }
      return 1;
   }
#ifdef DEBUG
  cerr << "STDIO: '" << locus << "' Not found in index. Tracking. iLastPos= " << iLastPos << "\n";
  int iCnt = 0;
#endif
   long iHP;
   long iDP;
   iHP = iLastPos;
   iDP = iLastPos;
   char sLine[MAXLINELENGTH];
   char sName[MAXLINELENGTH];
   if(ifs != 0) {
      fseek(ifs, iLastPos, SEEK_SET);
      clearerr(ifs);
      while(fgets(sLine, MAXLINELENGTH, ifs) != 0) {
         if(sLine[0] == ';' || sLine[0] == '>') {
            switch(sLine[0]) {
               case '>':
                  sscanf(sLine + 1, "%s", sName);
                  break;
               case ';':
                  while(fgets(sLine, MAXLINELENGTH, ifs) != 0 && sLine[0] == ';' ) {
                     sLine[0] = '\0';
                  }
                  sscanf(sLine, "%s", sName);
                  break;
            }
            iLastPos = iDP = ftell(ifs);
            lpmap[sName].assign(sDbName, iHP, iDP);
#ifdef DEBUG
            iCnt++;
            cerr << sName << "<=>" << locus << "\n";
#endif
            if(strcmp(sName, locus) == 0) {
               return 1;
            }
         }
         iHP = iLastPos = ftell(ifs);
      }
#ifdef DEBUG
      cerr << "Scanned " << iCnt << " locies\n";
#endif
   }
   return 0;
}
#endif
/** ==================================================== */
int CDb::loadIdx(const char *fn) {
   FILE *fidx;
   char sLine[MAXLINELENGTH];
   char sLName[MAXNAMESIZE];
   char sFName[MAXNAMESIZE];
   long nCom = 0, nFrom = 0;
   lpmap.clear();
   if((fidx = fopen(fn, "r")) == 0) {
      return 0;
   }
   while(fgets(sLine, MAXLINELENGTH, fidx) != 0) {
      if(sscanf(sLine, "%s%s%ld%ld", sFName, sLName, &nCom, &nFrom) < 4) {
         lpmap.clear();
         break;
      }
      lpmap[sLName].assign(sFName, nCom, nFrom);
   }
   fclose(fidx);
   return lpmap.size();
}
/** ==================================================== */
void CDb::saveIdx(const char* fn){
   FILE* fidx;
   if((fidx = fopen(fn, "w"))) {
     saveIdx(fidx);
     fclose(fidx);
   }
}
/** ==================================================== */
void CDb::saveIdx(FILE* fidx){
  LPMAP::iterator mptr;
  for(mptr = lpmap.begin(); mptr != lpmap.end(); mptr++) {
    fprintf(fidx, "%s\t%s\t%ld\t%ld\n",
       (*mptr).second.fname.c_str(), (*mptr).first.c_str(), (*mptr).second.headPos, (*mptr).second.dataPos);
    }
}
/** ==================================================== */
int CDb::open_ifs(const char *fn){
   if(fn == 0) {
      sDbName.assign("");
      ifs = 0;
      fprintf(stderr, "CDb::open_ifs(const char*): empty file name\n");
      return 0;
   }
   string fname;
   if(sDbName.c_str() != fn) {
       sDbName.assign(fn);
   }
   fname.assign("");
//   if(!exist(sDbName.c_str())) {
     fname = sDbDir;
//   }
   fname.append(sDbName);
#ifdef IOSTREAM
   ifs->clear();
   ifs->open(fname.c_str());
   if(!ifs->is_open()) {
#else
   ifs = fopen(fname.c_str(), "r");
   if(ifs == 0) {
#endif
      sDbName.assign("");
      fprintf(stderr, "Failed to open '%s' for reading\n", fname.c_str());
      return 0;
   }
   sDbName = fn;
#ifdef DEBUG
   cerr << "Opened " << sDbName << " as " << fname << "\n";
#endif
   return 1;
}
/** ==================================================== */
void CDb::close_ifs(){
#ifdef DEBUG
   cerr << "Closing " << sDbName << "\n";
#endif
   sDbName.assign("");
#ifdef IOSTREAM
   if(ifs->is_open()) {
      ifs->close();
   }
#else
   if(ifs != 0) {
      fclose(ifs);
      ifs = 0;
   }
#endif
}
/** ==================================================== */
void CDb::rewind_ifs(){
#ifdef IOSTREAM
   if(ifs->is_open()) {
      ifs->seekg(0, ios::beg);
   }
#else
   if(ifs != 0) {
      rewind(ifs);
      ifs = 0;
   }
#endif
}
/** ==================================================== */
bool CDb::exist(const char *pcFName) {
  FILE *FIN;
  if((FIN = fopen(pcFName, "r")) == 0) {
    return false;
  }
  fclose(FIN);
  return true;
}
/** ==============================================================
Reads a fragment of specified locus into sq
Returns number of symbols actually written into sq
============================================================== */
long CDb::fragment(const char* pcName, long iF, long iT, BioSubSequence& sq, bool bGap){
   if(iF <= 0) {
      iF = 1;
   }
   if(iT <= 0) {
      iT = 1;
   }

   sq.clear();
   //Reading header of the source sequence to get it's position if available.
   sq.readHead(getHead(pcName));
   if(getData(pcName) == 0) {
      fprintf(stderr, "CDb::fragment(const char*, long, long, BioSubSequence&): %s not found\n", pcName);
      return 0;
   }
   if(iF > iT) {
        sq.adjust_position(iT, iF - fragment(iT, iF, sq, bGap));
        sq.complement();
   }
   else {
        sq.adjust_position(iF, iT - fragment(iF, iT, sq, bGap));
   }
   return sq.size();
}
/** ==============================================================
Calculaes starting position of the fragment and reads it.
Returns difference between what should be read and what actually was read.
============================================================== */
long CDb::fragment(long iFr, long iTo, string& sq, bool bGap){
   long iF = min(iFr, iTo);
   long iT = max(iFr, iTo);
   long iW = getWidth();
   if(iW == 0) {
      fprintf(stderr, "CDb::getWidth() returned 0\n");
      return iT - iF + 1;
   }
   int iRP = iF + (iF - 1)/iW - 1;
#ifdef IOSTREAM
   ifs->seekg(iRP, ios_base::cur);
#else
   fseek(ifs, iRP, SEEK_CUR);
#endif
   return fragment(ifs, iT - iF + 1, sq, bGap);
}
/** ==============================================================
This function reads iCnt letters from ifs to sq, or untill ';' or '>' is met
Returns difference between what should be read and what actually was read.
============================================================== */
long CDb::fragment(ifstream *ifs, long iCnt, string& sq, bool bGap){
   char sLine[MAXLINELENGTH];
   ifs->clear();
   sq.assign("");
   while(ifs->getline(sLine, MAXLINELENGTH).good() && sLine[0] != ';' && sLine[0] != '>') {
      for(char *pcLine = sLine; *pcLine != '\n' && *pcLine != '\0'; pcLine++) {
         if((!bGap && isalpha(*pcLine)) || (bGap && isgraph(*pcLine))) {
            sq.append(1, *pcLine);
            if(--iCnt <= 0) {
               return 0;
            }
         }
      }
   }
   return iCnt;
}
/** ==============================================================
This function reads iCnt letters from ifs to sq, or untill ';' or '>' is met
Returns difference between what should be read and what actually was read.
============================================================== */
long CDb::fragment(FILE *ifs, long iCnt, string& sq, bool bGap){
   char sLine[MAXLINELENGTH];
   sq.assign("");
   while(fgets(sLine, MAXLINELENGTH, ifs) != 0 && sLine[0] != ';' && sLine[0] != '>') {
      for(char *pcLine = sLine; *pcLine != '\n' && *pcLine != '\0'; pcLine++) {
         if((!bGap && isalpha(*pcLine)) || (bGap && isgraph(*pcLine))) {
            sq.append(1, *pcLine);
            if(--iCnt <= 0) {
               return 0;
            }
         }
      }
   }
   return iCnt;
}
/** ==============================================================
Estimates length of data lines in sequence file.
The whole algorithm assumes that all lines in the sequence file are of the same length
Returns astimated length of datalines (width of the file)
============================================================= */
long CDb::getWidth(){
#ifdef IOSTREAM
   long iPos = ifs->tellg();
   ifs->clear();
#else
   long iPos = ftell(ifs);
#endif
   long iWf = 0;
   long iWt = 0;
   char c;
#ifdef IOSTREAM
   while(ifs->get(c).good()) {
#else
   while((c = fgetc(ifs)) != EOF) {
#endif
			if(c == ';' || c == '>') {
				 break;
			}
      if(c == '\n' && iWt > 0) {
         if(iWf == iWt) {
            break;
         }
         if(iWf == 0) {
            iWf = iWt;
            iWt = 0;
            continue;
         }
         if(iWf != iWt) {
            //fprintf(stderr, "CDb::getWidth(): First 2 lines are of different length: %ld - %ld\n", iWf, iWt);
            //fprintf(stderr, "Possible reasons: short sequence (less than 2 lines), outdated index or bad sequence format.\n");
         }
         if(iWf < iWt) {
            iWf = iWt;
            break;
         }
      }
      else {
         iWt++;
      }
   }
#ifdef IOSTREAM
   ifs->seekg(iPos, ios::beg);
   ifs->clear();
#else
   fseek(ifs, iPos, SEEK_SET);
#endif
   return iWf;
}
/** No descriptions */
int CDb::loadIndexUpToDate(const char* db_name, const char* idx_name) {
   if(db_name == 0) {
      return 0;
   }
   if(cmpage(db_name, idx_name) > 0) {
      fprintf(stderr, "Index file %s is older than %s\n", idx_name, db_name);
      return 0;
   }
   return loadIdx(idx_name);
}
/** No descriptions */
int cmpage(const char* name1, const char* name2) {
   if(name1 == 0 || name2 == 0) {
      return 0;
   }
   struct stat stat1;
   struct stat stat2;
   if(stat(name1, &stat1)) {
#ifdef DEBUG
     cerr << "::stat(" << name1 << ", &stat1); failed\n";
#endif
      return 0;
   }
   if(stat(name2, &stat2)) {
#ifdef DEBUG
     cerr << "::stat(" << name2 << ", &stat2); failed\n";
#endif
      return 0;
   }
#ifdef DEBUG
  cerr << name1 << " : mtime = " << stat1.st_mtime << " : ctime = " << stat1.st_ctime << " : atime = " << stat1.st_atime << "\n";
  cerr << name2 << " : mtime = " << stat2.st_mtime << " : ctime = " << stat2.st_ctime << " : atime = " << stat2.st_atime << "\n";
#endif
   return stat1.st_mtime - stat2.st_mtime;
}
